from media_toolkit.utils.dependency_requirements import requires_numpy, requires_cv2, requires
from media_toolkit.core.media_file import MediaFile

try:
    import cv2
    import numpy as np
except ImportError:
    pass


class ImageFile(MediaFile):
    """
    Has file conversions that make it easy to work with image files across the web.
    Internally it uses cv2 file format.
    """
    @requires('cv2', 'numpy')
    def from_np_array(self, np_array, img_type: str = None):
        if isinstance(np_array, list):
            np_array = np.array(np_array)

        if img_type is None:
            if "image/" not in self.content_type:
                img_type, self._channels = self.detect_image_type_and_channels(np_array)
            else:
                img_type = self.content_type.split("/")[1]
            self.content_type = f"image/{img_type}"

        is_success, buffer = cv2.imencode(f".{img_type}", np_array)
        if is_success:
            # avoid to check again for image type by calling super().from_bytes instead of self.from_bytes
            return super().from_bytes(buffer)
        else:
            raise ValueError(f"Could not convert np_array to {img_type} image")


    @requires('numpy', 'cv2')
    def to_np_array(self):
        bytes = self.to_bytes()
        return cv2.imdecode(np.frombuffer(bytes, np.uint8), -1)

    @requires_numpy()
    def to_cv2_img(self):
        return self.to_np_array()

    @requires_cv2()
    def save(self, path: str):
        cv2.imwrite(path, self.to_np_array())

    def _file_info(self):
        super()._file_info()
        np_array = self.to_np_array()
        img_type, self._channels = self.detect_image_type_and_channels(np_array)
        if img_type is not None:
            self.content_type = f"image/{img_type}"
        

    @staticmethod
    def detect_image_type_and_channels(image) -> (str, int):
        """Detect the image type and number of _channels from a numpy array."""
        if isinstance(image, list):
            image = np.array(image)

        # Check the number of _channels
        if len(image.shape) == 2:
            channels = 1  # Grayscale
        elif len(image.shape) == 3:
            channels = image.shape[2]
        else:
            #raise ValueError("Unsupported image shape: {}".format(image.shape))
            return None, None

        # Detect image type by checking for specific markers
        image_type = None

        # Convert to bytes and inspect file signatures for format detection
        success, encoded_image = cv2.imencode('.png', image)
        if success:
            encoded_bytes = encoded_image.tobytes()
            if encoded_bytes.startswith(b'\x89PNG\r\n\x1a\n'):
                image_type = 'png'
            elif encoded_bytes[0:2] == b'\xff\xd8':
                image_type = 'jpeg'
            elif encoded_bytes.startswith(b'BM'):
                image_type = 'bmp'
            elif encoded_bytes.startswith(b'GIF'):
                image_type = 'gif'

        return image_type, channels




class SimpleGeneratorWrapper:
    """
    Transforms a generator function into an iterable object with a length property.
    This is useful for example in tqdm
    """
    def __init__(self, generator, length):
        self.generator = generator
        self.length = length

    def __iter__(self):
        return self.generator

    def __len__(self):
        return self.length



import functools
from importlib.util import find_spec
from typing import Union

# this is a cache to speed up our checks
_installed_libs = []


def requirement_decorator(requirement: str):
    """
    Decorator to require a library.
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # did we already check for the requirement
            if requirement in _installed_libs:
                return func(*args, **kwargs)

            # find in modules
            spec = find_spec(requirement)
            if spec is None:
                raise ImportError(f"{requirement} is not installed. Please install {requirement} to use this function.")
            _installed_libs.append(requirement)

            return func(*args, **kwargs)
        return wrapper
    return decorator


def requires_numpy():
    return requirement_decorator("numpy")
def requires_cv2():
    return requirement_decorator("cv2")

def requires(requirements: Union[list, tuple, str], *args):
    """
    Decorator to require multiple libraries.
    """
    if isinstance(requirements, str):
        requirements = [requirements]
    elif isinstance(requirements, tuple):
        requirements = list(requirements)

    for arg in args:
        requirements.append(arg)

    req_func = lambda x: x  # identity function if no reqs given
    for req in requirements:
        req_func = requirement_decorator(req)

    return req_func




from media_toolkit.utils.dependency_requirements import requires
from media_toolkit.core.media_file import MediaFile
import io

from media_toolkit.utils.generator_wrapper import SimpleGeneratorWrapper

try:
    import soundfile
    import numpy as np
except ImportError:
    pass


class AudioFile(MediaFile):
    """
    Has file conversions that make it easy to work with image files across the web.
    """
    @requires('soundfile')
    def to_soundfile(self):
        return soundfile.read(self._content_buffer)

    @requires('soundfile')
    def to_np_array(self, sr: int = None, return_sample_rate: bool = False):
        self._content_buffer.seek(0)
        audio, sr = soundfile.read(self._content_buffer, samplerate=sr)  # sr=None returns the native sample rate
        if return_sample_rate:
            return audio, sr
        return audio

    @requires('soundfile')
    def from_np_array(self, np_array, sr: int = 44100, file_type: str = "wav"):
        sr = 44100 if sr is None else sr  # 22050
        # write to virtual file with librosa
        virtual_file = io.BytesIO()
        virtual_file.name = f"audio_file.{file_type}"
        soundfile.write(virtual_file, np_array, samplerate=sr, format=file_type)
        return super().from_file(virtual_file)

    @requires('soundfile')
    def to_stream(self, chunks_per_second: int = 10):
        """
        Generator that yields audio chunks of 1/chunks_per_second as numpy arrays.
        :param chunks_per_second: Number of chunks per second
        :return: Generator that yields numpy arrays with audio data.
        """

        audio, sample_rate = self.to_soundfile()
        chunk_size = sample_rate // chunks_per_second
        n_chunks = len(audio) // chunk_size + 1

        def generator():
            for i in range(n_chunks):
                chunk = audio[i * chunk_size: i * chunk_size + chunk_size]
                yield chunk.astype(np.float32) #.tobytes()

        g = SimpleGeneratorWrapper(generator=generator(), length=n_chunks)
        g.sample_rate = sample_rate
        return g



import tempfile
from typing import Union
import tqdm

from media_toolkit.utils.dependency_requirements import requires
import subprocess
import os

try:
    from pydub import AudioSegment
    from pydub.utils import mediainfo
except ImportError:
    pass

try:
    import numpy as np
except ImportError:
    pass

try:
    from vidgear.gears import WriteGear
except ImportError:
    pass

try:
    import cv2
except ImportError:
    pass


@requires('pydub', 'numpy')
def add_audio_to_video_file(
        video_file: str,
        audio_file: Union[str, list],
        save_path: str = None
):
    """
    Adds audio_file to a video file and saves it it to save_path. If save_path is None, a tempfile is created.
    :return: The path to the video file.
    """
    # convert to abs file paths
    video_file = os.path.abspath(video_file)
    audio_file = os.path.abspath(audio_file)
    if save_path is None:
        save_path = tempfile.NamedTemporaryFile(delete=False, suffix=".mp4").name
    if save_path == video_file:
        save_path = save_path.replace(".mp4", "_with_audio.mp4")

    try:
        # https://stackoverflow.com/questions/11779490/how-to-add-a-new-audio-not-mixing-into-a-video-using-ffmpeg
        # os.system(f"ffmpeg -i {video_file} -i {audio_file} -map 0:v -map 1:a -c:v copy -shortest {save_path}.mp4")
        p = subprocess.Popen((
            "ffmpeg",
            "-y", "-i", video_file, "-i", audio_file, "-map", "0:v", "-map", "1:a", "-c:v", "copy",
            "-shortest", save_path
        ))
        p.wait()
    except Exception as e:
        print(f"Error adding audio_file to video: {e}", e.__traceback__)

    return save_path


@requires('pydub', 'numpy')
def audio_array_to_audio_file(audio_array, sample_rate: int = 44100, save_path: str = None, audio_format: str = "mp3") -> str:
    """
    Saves an audio array to an audio file.
    :param audio_array: A numpy array containing the audio samples.
        Can be 1D or 2D (stereo). In form np.array([[array_frame_1], [array_frame_2], ..])
    """
    # audio_array in fom numpy to audio_file file saved in temporary file
    audio_array = np.array(audio_array, dtype=np.int16)

    channels = 2 if audio_array.ndim == 2 else 1
    song = AudioSegment(
        audio_array.tobytes(),
        frame_rate=sample_rate,
        sample_width=audio_array.dtype.itemsize,
        channels=channels
    )
    if save_path is None:
        temp_audio_file = tempfile.NamedTemporaryFile(delete=False, suffix=f".{audio_format}")
        save_path = temp_audio_file.name
    song.export(save_path, format=audio_format)
    return save_path


@requires('pydub')
def get_audio_sample_rate_from_file(file_path: str) -> int:
    info = mediainfo(file_path)
    if "sample_rate" not in info:
        raise ValueError("The audio file does not have a sample rate.")

    return int(info['sample_rate'])


@requires('vidgear', 'numpy', 'cv2')
def video_from_image_generator(image_generator, save_path: str = None, frame_rate: int = 30):
    """
    Creates a video from an image generator. The image generator should yield images as numpy arrays or filepaths.
    Returns a path to a tempfile if save_path is None, otherwise saves the video to the save_path.
        Don't forget to delete the tempfile later.
    :param image_generator: A generator or iterable that yields images as numpy arrays or filepaths.
    """
    # if save_path is None, return a VideoFile object
    if save_path is None:
        tempf = tempfile.NamedTemporaryFile(delete=False, suffix=".mp4")
        save_path = tempf.name

    # make a nice progress bar
    if hasattr(image_generator, "__len__"):
        image_generator = tqdm.tqdm(enumerate(image_generator), total=len(image_generator))
    else:
        image_generator = tqdm.tqdm(enumerate(image_generator))

    # Write the video
    output_params = {"-input_framerate": frame_rate}  # ffmpeg params
    writer = WriteGear(output=save_path, compression_mode=True, logging=True, **output_params)

    for i, img in image_generator:
        try:
            if isinstance(img, str):
                img = cv2.imread(img)
            if not isinstance(img, np.ndarray):
                raise ValueError("The image generator should yield images as numpy arrays or filepaths.")
            writer.write(img)
        except Exception as e:
            file_name = img if isinstance(img, str) else f"image_{i}"
            print(f"Error reading {file_name}: {e}. Skipping frame {i}")
            continue
    # Safely close the writer
    writer.close()
    return save_path





import glob
import os
import tempfile
from io import BytesIO
from typing import List, Union

from media_toolkit.core.video.video_utils import (add_audio_to_video_file, audio_array_to_audio_file,
                                                  video_from_image_generator, get_audio_sample_rate_from_file)
from media_toolkit.utils.generator_wrapper import SimpleGeneratorWrapper
from media_toolkit.utils.dependency_requirements import requires
from media_toolkit.core.media_file import MediaFile

try:
    import cv2
    import numpy as np
except ImportError:
    pass

try:
    from vidgear.gears import VideoGear, WriteGear
except:
    pass

try:
    from pydub import AudioSegment
    from pydub.utils import mediainfo
except ImportError:
    pass


class VideoFile(MediaFile):
    """
    A class to represent a video file.
    """

    def __init__(self):
        super().__init__()
        self.content_type = "video"
        self.frame_count = None
        self.frame_rate = None
        self.width = None
        self.height = None
        self.shape = None
        self.duration = None
        self.audio_sample_rate = None

    def from_files(self, image_files: Union[List[str], list], frame_rate: int = 30, audio_file=None):
        """
        Creates a video based of a list of image files and an audio_file file.
        :param image_files: A list of image files to convert to a video. Either paths or numpy arrays.
        :param frame_rate: The frame rate of the video.
        :param audio_file: The audio_file file to add to the video, as path, bytes or AudioSegment.
        """
        # Check if there are images in the list
        if not image_files:
            raise ValueError("The list of image files is empty.")

        # Create a temporary file to store the video
        temp_vid_file_path = video_from_image_generator(image_files, frame_rate=frame_rate, save_path=None)
        # Merge video and audio_file using pydub
        if audio_file is not None:
            combined = add_audio_to_video_file(video_file=temp_vid_file_path, audio_file=audio_file)
            self.from_file(combined)
            os.remove(combined)
            os.remove(temp_vid_file_path)
            return self

        # Init self from the temp file
        self.from_file(temp_vid_file_path)
        # remove tempfile
        os.remove(temp_vid_file_path)

        return self

    def from_image_files(self, image_files: List[str], frame_rate: int = 30):
        """
        Converts a list of image files into a video file.
        """
        return self.from_files(image_files, frame_rate, audio_file=None)

    def from_dir(self, dir_path: str, audio: Union[str, list] = None, frame_rate: int = 30):
        """
        Converts all images in a directory into a video file.
        """
        image_types = ["*.png", "*.jpg", "*.jpeg"]
        image_files = []
        for image_type in image_types:
            image_files.extend(glob.glob(os.path.join(dir_path, image_type)))
        # sort by date to make sure the order is correct
        image_files.sort(key=lambda x: os.path.getmtime(x))

        # if audio_file is none, take the first audio_file file in the directory
        if audio is None:
            audio_types = ["*.wav", "*.mp3"]
            for audio_type in audio_types:
                audio = glob.glob(os.path.join(dir_path, audio_type))
                if len(audio) > 0:
                    audio = audio[0]
                    break

        return self.from_files(image_files=image_files, frame_rate=frame_rate, audio_file=audio)

    def add_audio(self, audio_file: Union[str, list], sample_rate: int = 44100):
        """
        Adds audio to the video file.
        :param audio_file: The audio_file file to add to the video, as path, or numpy array.
            In case of a file, the sample rate is determined from the file.
        :param sample_rate: If the audio_file is a numpy array, the sample rate should be provided.
        """

        if self.audio_sample_rate is None:
            if self.frame_rate is None:
                raise Exception("The frame rate of the video file is not set. Read a video file first.")

            if os.path.isfile(audio_file):
                self.audio_sample_rate = get_audio_sample_rate_from_file(audio_file)
            else:
                self.audio_sample_rate = get_audio_sample_rate_from_file(self._to_temp_file())

        if isinstance(audio_file, list) or isinstance(audio_file, np.ndarray):
            audio_file = audio_array_to_audio_file(audio_file, sample_rate=sample_rate)

        tmp = self._to_temp_file()
        combined = add_audio_to_video_file(tmp, audio_file)
        self.from_file(combined)
        os.remove(tmp)
        os.remove(combined)
        return self

    def read(self):
        self._content_buffer.seek(0)
        return self._content_buffer.getvalue()

    def _to_temp_file(self):
        # get suffix
        if self.content_type is None:
            raise ValueError("The content type of the video file is not set.")
        if "/" not in self.content_type:
            raise ValueError("The content type of the video file is not valid. Read a video file first.")
        suffix = self.content_type.split("/")[1]
        if suffix == 'octet-stream':
            raise ValueError("The content type of the video file is not valid. Read a video file first.")

        with tempfile.NamedTemporaryFile(delete=False, suffix=f".{suffix}") as temp_video_file:
            temp_video_file.write(self.read())
            temp_video_file_path = temp_video_file.name

        return temp_video_file_path

    @requires('vidgear', 'numpy', 'pydub')
    def from_video_stream(self, video_audio_stream, frame_rate: int = 30, audio_sample_rate: int = 44100):
        """
        Given a generator that yields video frames and audio_file data as numpy arrays, this creates a video.
        The generator is expected to be in the form of: VideoFile().to_video_stream()
            or a generator that yields images as numpy arrays like VideoFile().to_image_stream().
        """
        # Reset and pre-settings
        self._reset_buffer()

        # new generator, to extract audio_file
        audio_frames = []

        def _frame_gen():
            for frame in video_audio_stream:
                # check if is video and audio_file stream or only video stream
                if isinstance(frame, tuple) and len(frame) == 2:
                    frame, audio_data = frame
                    if audio_data is None:
                        audio_data = np.zeros(0)
                        print("Warning: Audio data is None. Adding silence in frame.")
                    audio_frames.append(audio_data)
                yield frame

        # allows tqdm to work with the generator
        video_gen_wrapper = _frame_gen()
        if hasattr(video_audio_stream, '__len__'):
            video_gen_wrapper = SimpleGeneratorWrapper(video_gen_wrapper, length=len(video_audio_stream))

        # Create video
        temp_video_file_path = video_from_image_generator(video_gen_wrapper, frame_rate=frame_rate, save_path=None)

        # Add audio_file
        if len(audio_frames) > 0:
            temp_audio_file = audio_array_to_audio_file(audio_frames, sample_rate=audio_sample_rate)
            combined = add_audio_to_video_file(temp_video_file_path, temp_audio_file)
            self.from_file(combined)
            # cleanup
            os.remove(temp_audio_file)
            os.remove(temp_video_file_path)
            os.remove(combined)
        else:
            self.from_file(temp_video_file_path)
            os.remove(temp_video_file_path)

        return self

    @requires('cv2', 'pydub')
    def _file_info(self):
        super()._file_info()  # sets: file_name, content_type.

        # care for the case that it was loaded from_bytes what usually does not provide any filename / info.
        # In this case we need to write the data first to file and then retrieve the info again.
        path = self.path
        is_temp_file = False
        if path is None or not os.path.exists(path):
            self._content_buffer.seek(0)
            path = self._to_temp_file()
            is_temp_file = True

        # get video info
        info = mediainfo(path)

        def info_to_number(key: str, default_val=None):
            if key in info:
                val = info[key]
                if val == 'N/A':
                    return default_val
                # split if / in val and take first
                val = val.split("/")[0]
                return float(val)
            return default_val

        self.frame_count = info_to_number('nb_frames')
        self.duration = info_to_number('duration')
        self.width = info_to_number('width')
        self.height = info_to_number('height')
        self.shape = (self.width, self.height)
        self.audio_sample_rate = info_to_number('sample_rate', 44100)

        self.frame_rate = info_to_number('avg_frame_rate', None)
        # need to determine the frame rate with cv2 because pydub calculation gives some weird results..
        if self.frame_rate is None or self.frame_count is None or self.frame_count == 1:
            cap = cv2.VideoCapture(path)
            self.frame_rate = cap.get(cv2.CAP_PROP_FPS)
            self.frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            cap.release()

        # if self.width is None or self.height is None:
        #    # try to get it with cv2
        #    cap = cv2.VideoCapture(path)
        #    self.width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        #    self.height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        #    self.shape = (self.width, self.height)
        #    self.frame_rate = cap.get(cv2.CAP_PROP_FPS)
        #    cap.release()
        # else:
        #    if self.frame_count is not None and self.duration is not None:
        #        self.frame_rate = int(self.frame_count / self.duration)

        if 'format_name' in info:
            format_name = info['format_name'].split(",")[0]
            self.content_type = f"video/{format_name}"
        else:
            self.content_type = "video/mp4"  # overwrite default "application/octet-stream"

        # if is tempfile remove it
        if is_temp_file:
            os.remove(path)

    @requires('vidgear')
    def to_image_stream(self):
        return self.to_video_stream(include_audio=False)

    @requires('pydub', 'vidgear')
    def to_video_stream(self, include_audio=True):
        """
        Yields video frames and audio_file data as numpy arrays.
        :param include_audio: if the audio_file is included in the video stream. If not it will only yield the video frames.
        :return:
        """
        self._content_buffer.seek(0)
        # because CamGear does not support reading from a BytesIO buffer, we need to save the buffer to a temporary file
        temp_video_file_path = self._to_temp_file()
        stream = VideoGear(source=temp_video_file_path).start()

        if include_audio:
            # Extract audio_file using pydub
            try:
                audio = AudioSegment.from_file(temp_video_file_path)
                # Calculate the audio_file segment duration per frame
                audio_per_frame_duration = 1000 / stream.framerate  # duration of each video frame in ms
            except:
                include_audio = False
                print("Could not extract audio_file from video file. Audio will not be included in the video stream.")

        # Initialize frame counter for audio_file and better self.frame count
        frame_count = 0
        audio_shape = None  # if the audio in a frame is to short or mishaped fill it with silence.
        try:
            while True:
                # Read frame
                frame = stream.read()
                if frame is None:
                    break

                if not include_audio:
                    yield frame
                    continue

                # Calculate the start and end times for the corresponding audio_file segment
                start_time = frame_count * audio_per_frame_duration
                end_time = start_time + audio_per_frame_duration
                frame_audio = audio[start_time:end_time]

                # Convert audio_file segment to raw data
                audio_data = np.array(frame_audio.get_array_of_samples())

                # CODE TO DEAL WITH ERRORS AND IMPUTE VALUES
                # save the first shape of the audio data
                if audio_shape is None and len(audio_data) > 0:
                    audio_shape = audio_data.shape

                if audio_data is None:
                    # sometimes in a frame theres no audio data. Then we need to fill it with silence.
                    audio_data = np.zeros(audio_shape)

                # impute missing values or cut too long audio arrays
                if audio_shape is not None:
                    if len(audio_data) < audio_shape[0]:
                        audio_data = np.pad(audio_data, (0, audio_shape[0] - len(audio_data)), 'constant')
                    elif len(audio_data) > audio_shape[0]:
                        audio_data = audio_data[:len(audio_shape)]

                # Yield the frame and the corresponding audio_file data
                yield frame, audio_data

                # Increment frame counter
                frame_count += 1
        finally:
            # Safely close the video stream
            stream.stop()
            # Remove the temporary video file
            os.remove(temp_video_file_path)
            # accurate value instead of using cv2.CAP_PROP_FRAME_COUNT
            self.frame_count = frame_count

    @requires('pydub')
    def extract_audio(self, path: str = None, export_type: str = 'mp3') -> Union[bytes, None]:
        temp_video_file_path = self._to_temp_file()
        audio = AudioSegment.from_file(temp_video_file_path)

        if path is not None and len(path) > 0:
            dirname = os.path.dirname(path)
            if len(dirname) > 0 and not os.path.isdir(dirname):
                os.makedirs(dirname)
            audio.export(path, format=export_type)
            os.remove(temp_video_file_path)
            return None

        # return as bytes
        file = BytesIO()
        file = audio.export(file, format=export_type)
        file.seek(0)
        data = file.read()
        file.close()
        # remove tempfile
        os.remove(temp_video_file_path)
        return data

    def __iter__(self):
        return self.to_video_stream()

    def __len__(self):
        return self.frame_count



import base64
import io
import mimetypes
from typing import Union, BinaryIO
import os
from media_toolkit.utils.dependency_requirements import requires_numpy


try:
    import numpy as np
except ImportError:
    pass


class MediaFile:
    """
    Has file conversions that make it easy to work standardized with files across the web and in the sdk.
    Works natively with bytesio, base64 and binary data.
    """
    def __init__(self, file_name: str = "file", content_type: str = "application/octet-stream"):
        """
        :param file_name: The name of the file. Note it is overwritten if you use from_file/from_starlette.
        :param content_type: The content type of the file. Note it is overwritten if you use from_file/from_starlette.
        """
        self.content_type = content_type
        self.file_name = file_name  # the name of the file also when specified in bytesio
        self.path = None    # the path of the file if it was provided. Is also indicator if file was loaded from file.
        self._content_buffer = io.BytesIO()

    def from_any(self, data):
        """
        Load a file from any supported data type. The file is loaded into the memory as bytes.
        """
        # it is already converted
        if isinstance(data, MediaFile):
            return data

        # conversion factory
        if type(data) in [io.BufferedReader, io.BytesIO]:
            self.from_bytesio_or_handle(data)
        elif isinstance(data, str):
            if self._is_valid_file_path(data):
                self.from_file(data)
            else:
                self.from_base64(data)
        elif isinstance(data, bytes):
            self.from_bytes(data)
        elif type(data).__name__ == 'ndarray':
            self.from_np_array(data)
        elif data.__module__ == 'starlette.datastructures' and type(data).__name__ == 'UploadFile':
            self.from_starlette_upload_file(data)

        return self

    def from_bytesio_or_handle(
        self,
        buffer: Union[io.BytesIO, BinaryIO, io.BufferedReader],
        copy: bool = True
    ):
        """
        Set the content of the file from a BytesIO or a file handle.
        :params buffer: The buffer to read from.
        :params copy: If true, the buffer is completely read to bytes and the bytes copied to this file.
            If false file works with the provided buffer. Danger -- The buffer is kept open.
        """
        if not type(buffer) in [io.BytesIO, io.BufferedReader]:
            raise ValueError(f"Buffer must be of type BytesIO or BufferedReader. Got {type(buffer)}")

        self._reset_buffer()
        buffer.seek(0)

        # setting path is needed in order that file_info can work properly
        if type(buffer) in [io.BufferedReader]:
            self.path = buffer.name

        if not copy:
            self._content_buffer = buffer
            self._file_info()
        else:
            self.from_bytes(buffer.read())  # calls self._file_info also
            buffer.seek(0)

        return self

    def from_bytesio(self, buffer: Union[io.BytesIO, BinaryIO], copy: bool = True):
        return self.from_bytesio_or_handle(buffer=buffer, copy=copy)

    #@staticmethod
    #@overload
    #def from_file(path_or_handle: Union[str, io.BytesIO, io.BufferedReader]):
    #    return MediaFile().from_file(path_or_handle)

    def from_file(self, path_or_handle: Union[str, io.BytesIO, io.BufferedReader]):
        """
        Load a file from a file path, file handle or base64 and convert it to BytesIO.
        """
        if type(path_or_handle) in [io.BufferedReader, io.BytesIO]:
            self.from_bytesio_or_handle(path_or_handle)
        elif isinstance(path_or_handle, str):
            # read file from path
            if not os.path.exists(path_or_handle):
                raise FileNotFoundError(f"File {path_or_handle} not found.")

            self.path = path_or_handle
            # self.content_type = mimetypes.guess_type(self.file_name)[0] or "application/octet-stream"
            with open(path_or_handle, 'rb') as file:
                self.from_bytesio_or_handle(file)  # method also calls self._file_info.

        return self

    def from_bytes(self, data: bytes):
        self._reset_buffer()
        self._content_buffer.write(data)
        self._content_buffer.seek(0)
        self._file_info()
        return self

    def from_starlette_upload_file(self, starlette_upload_file):
        """
        Load a file from a starlette upload file.
        :param starlette_upload_file:
        :return:
        """
        # in file info all the meta is retrieved from the file.name in case of buffered reader
        content = starlette_upload_file.file.read()
        self.file_name = starlette_upload_file.filename
        self.content_type = starlette_upload_file.content_type
        self.from_bytes(content)
        return self

    def from_base64(self, base64_str: str):
        """
        Load a file which was encoded as a base64 string.
        """

        decoded = self._decode_base_64_if_is(base64_str)
        if decoded is not None:
            return self.from_bytes(base64.b64decode(base64_str))
        else:
            err_str = base64_str if len(base64_str) <= 50 else base64_str[:50] + "..."
            raise ValueError(f"Decoding from base64 like string {err_str} was not possible. Check your data.")

    @requires_numpy()
    def from_np_array(self, np_array: np.array):
        """
        Convert a numpy array to a file which is saved as bytes b"\x93NUMPY" into the buffer.
        """
        self._reset_buffer()
        np.save(self._content_buffer, np_array)
        return self

    def from_dict(self, file_result_json: dict):
        """
        Load a file from a dictionary.
        :param d: The dictionary to load from formatted as FileResult.to_json().
        """
        self.file_name = file_result_json["file_name"]
        self.content_type = file_result_json["content_type"]
        # ToDo: the from_base64 might overwrite name and content type (ImageFile). Check if this always is intended.
        self.from_base64(file_result_json["content"])
        return self

    @requires_numpy()
    def to_np_array(self, shape=None, dtype=np.uint8):
        """
        If file was created with from_np_array it will return the numpy array.
        Else it will try to convert the file to a numpy array (note this is converted bytes representation of the file).
        :param shape: The shape of the numpy array. If None it will be returned flat.
        :param dtype: The dtype of the numpy array. If None it will be uint8.
        """
        bytes = self.to_bytes()
        # check if was saved with np.save so bytes contains NUMPY
        if bytes.startswith(b"\x93NUMPY"):
            self._content_buffer.seek(0)
            return np.load(self._content_buffer, allow_pickle=False)

        shape = shape or (1, len(bytes))
        dtype = dtype or np.uint8

        arr_flat = np.frombuffer(bytes, dtype=dtype)
        return arr_flat.reshape(shape)

    def to_bytes(self) -> bytes:
        return self.read()

    def read(self) -> bytes:
        self._content_buffer.seek(0)
        res = self._content_buffer.read()
        self._content_buffer.seek(0)
        return res

    def to_bytes_io(self) -> io.BytesIO:
        self._content_buffer.seek(0)
        return self._content_buffer

    def to_base64(self):
        return base64.b64encode(self.to_bytes()).decode()

    def to_httpx_send_able_tuple(self):
        return self.file_name, self.read(), self.content_type

    def _reset_buffer(self):
        self._content_buffer.seek(0)
        self._content_buffer.truncate(0)

    def save(self, path: str):
        """
        Methods saves the file to disk.
        If path is a folder it will save it in folder/self.filename.
        If path is a file it will save it there.
        :param path:
        :return:
        """
        # create folder if not exists
        if os.path.dirname(path) != "" and not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))
        # check if path contains a file name
        if os.path.basename(path) == "":
            path = os.path.join(path, self.file_name)

        with open(path, 'wb') as file:
            file.write(self.read())

    def _file_info(self):
        """
        After writing the file to the buffer, this method is called to determine additional file informations.
        For videos this might be length, frame rate...
        If you subclass don't forget to call super()._file_info() to set the file name and content type.
        """
        # cases when file_info is called
        # from_file -> retrieve info directly from the file path
        # from bytesio -> tempfile
        # from bytes -> tempfile
        # from buffered_reader -> set path -> from bytes -> get info from previously set file_path
        # from np_array -> tempfile
        # from starlette_upload_file -> from_buffered_reader(spooled_temporary) -> info from the spooled_temporary
        # from base64 -> from-bytes -> tempfile
        if self.path is not None:
            self.file_name = os.path.basename(self.path)
            self.content_type = mimetypes.guess_type(self.file_name)[0] or "application/octet-stream"
        elif hasattr(self._content_buffer, "name"):
            self.file_name = os.path.basename(self._content_buffer.name)

        if self.content_type is None:
            self.content_type = "application/octet-stream"

    def __bytes__(self):
        return self.to_bytes()

    def __array__(self):
        return self.to_np_array()

    def to_json(self):
        """
        Returns the file as a json serializable dictionary.
        :return: { "file_name": str, "content_type": str, "content": str }
        """
        return {
            "file_name": self.file_name,
            "content_type": self.content_type,
            "content": self.to_base64()
        }

    @staticmethod
    def _decode_base_64_if_is(data: Union[bytes, str]):
        """
        Checks if a string is base64. If it is, it returns the base64 string as bytes; else returns None.
        """
        if isinstance(data, str):
            data = data.encode()

        try:
            # Decode the data
            decoded = base64.b64decode(data, validate=True)
            # Re-encode the decoded data
            back_encoded = base64.b64encode(decoded)
            # Compare with the original encoded data
            if back_encoded == data:
                return decoded
        except Exception:
            pass

        return None

    @staticmethod
    def _is_valid_file_path(path: str):
        try:
            is_file = os.path.isfile(path)
            return is_file
        except:
            return False




"""
Tests some core functionalities of the VideoFile class.
"""
from tqdm import tqdm

from media_toolkit import VideoFile
import cv2

outdir = "outdir/"

# test video files
#test_video = "test_files/test_video.mp4"
#vf = VideoFile().from_file(test_video)
#### extract audio_file
#vf.extract_audio(f"{outdir}/extracted_audio.mp3")
##audio_bytes = vf.extract_audio()
#
### test video clients
#audio_array = []
#image_paths = []
#for i, (img, audio_part) in tqdm(enumerate(vf.to_video_stream(include_audio=True))):
#    p = f"{outdir}/test_out_video_stream_{i}.png"
#    image_paths.append(p)
#    cv2.imwrite(p, img)
#    audio_array.append(audio_part)

# new video File
#vf = VideoFile().from_files(image_paths)
#vf.add_audio(audio_array)
#vf.save(f"{outdir}/test_from_files_add_audio.mp4")
# from dir; and combine audio and video
# files = [f"{outdir}/test_out_video_stream_{i}.png" for i in range(10)]

fromdir = VideoFile().from_dir(outdir, audio=f"{outdir}/extracted_audio.mp3", frame_rate=30)
fromdir.save(f"{outdir}/test_from_dir.mp4")

# test video clients with audio_file
fromstream = VideoFile().from_video_stream(fromdir.to_video_stream(include_audio=True))
fromstream.save(f"{outdir}/test_from_stream.mp4")


from media_toolkit import MediaFile, ImageFile, AudioFile, VideoFile


def convert_to_upload_file_type(file, target_media_file=None):
    """
    Converts a file to a send able format.
    :param file: The file to convert.
    :param target_media_file: The target type to convert to. If not specified will be converted to MediaFile.
        Use ImageFile, AudioFile, VideoFile to convert to those types.
    :return: The send able file.
    """
    # it is already converted
    if isinstance(file, MediaFile):
        return file

    # determine target class
    target_class = MediaFile
    if target_media_file is not None and issubclass(target_media_file, MediaFile):
        target_class = target_media_file
    media_file_instance = target_class()

    # load data
    media_file_instance = media_file_instance.from_any(file)
    return media_file_instance


def from_file_result(file_result: dict):
    """
    Converts a file result to a MediaFile.
    :param file_result: The file result to convert.
    :return: The MediaFile.
    """
    content_type = file_result.get("content_type", None)
    if content_type is None:
        content = file_result.get('content', file_result)
        return MediaFile().from_any(content)

    # determine target class from file type
    content_type = content_type.lower()
    file_types = {
        "octet-stream": MediaFile,
        "image": ImageFile,
        "audio_file": AudioFile,
        "video": VideoFile
    }
    # fancy way to write efficient the factory conversion
    target_class = next(
        filter(lambda ft: ft[0] in content_type, file_types.items()),
        (None, MediaFile) # return tuple as default because next returns key, value
    )[1]

    return target_class().from_dict(file_result)

